/*---------------------------------------------------------------------------------------
PSgraf_p.h: private part of header file for PSgraf.

history:
  simple 3d objects added; introduce struct dvec; C++ (3.0.0)                   031115kr
  encapsulated bitmap preview (2.4.1)                                           011226kr
  arbitrary triangular grids (2.4.0a)                                           001216kr
  added bitmaps and TIFF                                                        971121kr
  delete magnification and sPaperCoord                                          950330kr
  make all coordinates positive	(compatibility with other OS)                   950330kr
  variable argument lists for gray/color                                        920819kr
  color added (2.0)                                                             920523kr
  created based on SciGraf for the Macintosh (1.0)                              910608kr
---------------------------------------------------------------------------------------*/

#ifdef COM_SYM_DEF
#define EXTERN
#else
#define EXTERN extern
#endif
/*------------------------------------------------------------------------ definitions */
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "PSgraf3.h"
#include "tiffio.h"
#include "tiff.h"

#define PSgraflibVersion 303
#define PSgraflibVersionString "V3.0.3"

#define PT_to_MM     2.83465      /* scaling [point] --> [mm]                          */
#define X_stretch    0.56         /* used to estimate the extent of a string...        */
#define Y_stretch    0.73         /* ...in x- and y-direction from curFSize            */
#define CHUNK     4096            /* elements allocated per request for more memory    */
#define Deg_to_Rad   0.017453293  /* pi/180                                            */
#define PS_PIX_SIZE  0.25         /* pixel size (used in dLegend for density plots)    */

/*-------------------------------------------------------------------------- variables */
/*----- filenames, filestatus, number of drawings -------------------------------------*/
EXTERN int PSinitialized;         /* flag for initialization of PSgraf                 */
EXTERN char *paper_name;          /* name of current paper                             */
EXTERN long BoundingBox;          /* position of bounding box in PostScript file       */
EXTERN long PaperShift;           /* position of shifting vector for positive coord.   */
EXTERN long TeXoff;               /* position of xoff & yoff in TeX file               */
EXTERN int PSdrwN;                /* # drawings                                        */
EXTERN int EPSencaps;             /* 0: no preview; 1: add preview (using system call) */

/*----- coordinate systems, current position, bounding box ----------------------------*/
EXTERN double Pa,Pb,Pc,Pd;        /* scaling-rotation matrix for paper coordinates     */
EXTERN double Px,Py;              /* translation vector for paper coordinates          */
EXTERN double Wlx,Wly,Wlz,        /* left end of window in world coordinates           */
              Plx,Ply,Plz,        /* left end of window in paper coordinates           */
              PWx,PWy,PWz;        /* scaling factors [Pr-Pl]/[Wr-Wl]                   */
EXTERN double cpx,cpy;            /* current position (Paper coordinates [mm])         */
EXTERN double Bxll,Byll,Bxur,Byur;/* lower left and upper right corner of bounding box */
EXTERN int z_set;		  /* third dimension active			       */

/*----- R^3 --> R^2 projection --------------------------------------------------------*/
EXTERN double R32x0,R32y0,R32z0,  /* origin of image plane in normal space             */
              R32x1,R32y1,R32z1,  /* x vector of image plane in normal space           */
              R32x2,R32y2,R32z2;  /* y vector of image plane in normal space           */
EXTERN struct dvec R32_view;      /* unit vector pointing from observer to origin      */

/*----- pen & color -------------------------------------------------------------------*/
EXTERN double Pthick;             /* pen thickness [Paper]                             */
EXTERN int ColorSpace;            /* G=0, RGB=2, HSB=4, CMYK=8                         */
EXTERN double S_c0,F_c0,          /* color component 0 for stroke and fill             */
                                  /*   (F_c0=NONE --> no filling)                      */
              S_c1,F_c1,          /* color component 1 for stroke and fill             */
              S_c2,F_c2,          /* color component 2 for stroke and fill             */
              S_c3,F_c3;          /* color component 3 for stroke and fill             */
                                  /*   (0 if not used)                                 */

/*----- axes --------------------------------------------------------------------------*/
EXTERN double xinterval[3],       /* intervals for i-th order ticks (i=0,..,2)         */
  yinterval[3],zinterval[3];      /*    "                                              */
EXTERN double xTiMaLe[3],         /* length of i-th order ticks                        */
  yTiMaLe[3],zTiMaLe[3];          /*    "                                              */
EXTERN int xprecision,            /* precision for lettering the axes                  */
  yprecision,zprecision;          /*    "                                              */
EXTERN int xauto,yauto,zauto;     /* autom. determination of intervals and precision   */

/*----- clipping ----------------------------------------------------------------------*/
EXTERN int clip_on;               /* clipping on                                       */
EXTERN double clip_xl,clip_xh,    /* clipping bounds in x-direction (PC)               */
  clip_yl,clip_yh;                /* clipping bounds in y-direction (PC)               */

/*----- arrows ------------------------------------------------------------------------*/
EXTERN double arr_lhead,          /* length of head                                    */
  arr_wtail,arr_whead;            /* width of tail and head                            */
EXTERN char arr_root;             /* root of arrow (T,C,H)                             */
EXTERN int arr_tilt;              /* true if annotation is to be tilted                */

/*----- data point --------------------------------------------------------------------*/
EXTERN int dp_sym;                /* symbol for data point                             */
EXTERN double dp_rsym;            /* radius of symbol                                  */
EXTERN double dp_bar;             /* bar length for errors                             */

/*----- fonts & format for numbers ----------------------------------------------------*/
EXTERN int TeXfonts;              /* use TeXfonts: 0: no, 1: yes, 2: no, but file open */
EXTERN char TeXstyle[127];        /* TeX commands to include with every text           */
EXTERN int lay_text,lay_fill;     /* counters for text and fill layers (for TeX only)  */
EXTERN char curFont[63];          /* name of current font                              */
EXTERN double curFSize;           /* current font size [points]                        */
EXTERN char hadj,vadj;            /* horiz.&vert. adjust.: hadj=[L,C,R], vadj=[T,C,B]  */
EXTERN double TextRotation;       /* angle relative to PaperCoordinates                */
EXTERN char nformat;              /* format for writing numbers                        */
EXTERN int precision;             /* precision for writing numbers                     */

/*----- file pointer ------------------------------------------------------------------*/
EXTERN FILE *PSout;               /* PostScript output                                 */
EXTERN FILE *TeXout;              /* TeX output                                        */

/*----- grid for contouring -----------------------------------------------------------*/
struct line {
  int id;                         /* line number                                       */
  double bx,by,ex,ey;             /* coordinates of beginning (bx,by) and end (ex,ey)  */
  int boundary;                   /* 1: belongs to boundary                            */
  int n[2];                       /* id of beginning (n[0]) and end (n[1]) node        */
  struct line *l[6];              /* pointers to neighboring lines
                                     (only 4 used for triangular grid)                 */
  double xCP,yCP;                 /* coord. of crossing point on this line (if any)    */
  int state;                      /* 0: no CP; 1: unused CP; -1: CP already used       */
};
struct bline {
  struct line *l;                 /* pointer to grid line                              */
  struct bline *path[2];          /* next (path[0]), prev. (path[1]) line on boundary  */
  int state[2];                   /* 0: node[i] not yet used for contourline           */
};
EXTERN int GT_;                   /* grid type 0: topologically regular quadrangular
                                               1: arbitrary triangular                 */
EXTERN struct line *L_;           /* pointer to set of grid lines                      */
EXTERN int NL_;                   /* # lines in grid                                   */
EXTERN struct bline *B_;          /* pointer to beginning of boundary                  */
EXTERN int NB_;                   /* # lines on boundary                               */
EXTERN int N_gn;                  /* # grid nodes                                      */

/*----- contour lines & density plot --------------------------------------------------*/
EXTERN int N_Ctr;                 /* # contours || color transfer points               */
EXTERN double *C_ctr;             /* pointer to contour values                         */
EXTERN struct color *C_col;       /* pointers to colors                                */
EXTERN struct line *first_on_p;   /* first line on polygon                             */
struct polygon {
  double *x,*y;                   /* pointers to arrays with x- and y-coord. of path   */
  int *boundary;                  /* pointer to boundary array (1: belongs to boundary) 
                                     *boundary==NULL if this feature is not used       */
  int N;                          /* # points in x and y                               */
  double bxll,byll,bxur,byur;     /* bounding box of polygon (in WC)                   */
  int ictr;                       /* index of polygon's contour level                  */
  int color;                      /* polygon's color (index to C_col, -1 --> white)    */
  struct line *testL;             /* test line to determine if polygon is a hole       */
};
EXTERN struct polygon *P_;        /* set of all contour lines (polygons)               */
EXTERN int NP_;                   /* # contour lines (polygons)                        */
EXTERN double Min_Pix;            /* minimal pixel size (don't draw smaller poly)      */
EXTERN int below_Pix;             /* #polygons below Min_Pix                           */

/*----- element coverage of arbitrary trianangular grid -------------------------------*/
struct Telement {
  int id;                         /* element number                                    */
  int n[3];                       /* node numbers                                      */
  double cg[2];                   /* coordinates of center of gravity                  */
  int l[3];                       /* line numbers                                      */
};
EXTERN struct Telement *T_;       /* set of all triangles                              */
EXTERN int NT_;                   /* # triangular elements                             */

/*----- contoured block ---------------------------------------------------------------*/
EXTERN int CB_Nx,CB_il,CB_iu,     /* #nodes, lower&upper bound of block in x-direction */
  CB_Ny,CB_jl,CB_ju,              /* #nodes, lower&upper bound of block in y-direction */
  CB_Nz,CB_kl,CB_ku;              /* #nodes, lower&upper bound of block in z-direction */
EXTERN double CB_x0,CB_dx,        /* origin and grid constant in x-direction           */
  CB_y0,CB_dy,                    /* origin and grid constant in y-direction           */
  CB_z0,CB_dz;                    /* origin and grid constant in z-direction           */
EXTERN int CB_outer;              /* contour outer skin of block                       */
EXTERN struct dvec CB_illum;     /* direction of illumination (dim=0 -> no shading)   */


/*------------------------------------------------------------------ private functions */
void dP2P(double *x,double *y,int N);
void dP2W(double *x,double *y,int N);
void dP3P(double *x,double *y,double *z,int N);
void dP3W(double *x,double *y,double *z,int N);
void linea(char CS,struct dvec x);
void liner(char CS,struct dvec x);
void liner(char CS,double x,double y);
void l2Pa(double x,double y);
void l2Wa(double x,double y);
void m2Pa(double x,double y);
void m2Wa(double x,double y);
void m3Pa(double x,double y,double z);
void m3Wa(double x,double y,double z);
void updateBB(double x,double y);


#undef COM_SYM_DEF
