/* Tobin Fricke's implementation of the
   Hoshen-Kopelman algorithm for
   cluster labeling.

   Copyright (c) September 9, 2000, by Tobin Fricke <tobin@splorg.org>
   Distributed under the terms of the GNU Public License.

   Modified 2002-03-09 Tobin Fricke
   Modified substantially 2004-04-21 by Tobin Fricke

   This program is written in the 1999 standard of the C language (C99).  Older C
   compilers will refuse to compile it.   You can use a C++ compiler, a C99 compiler,
   or you can modify this code to comply with a previous version of the C standard.
   The GCC compiler supports C99 as of version 3.0.  Compile this program with:

   g++ -Wall hk.c -o hk -lQuantim4 -ltiff

   http://www.ocf.berkeley.edu/~fricke/projects/hoshenkopelman/hoshenkopelman.html
*/

#define MAX(a,b) (a>b?a:b)

#include "quantim4.h"
#include <assert.h>

int hoshen_kopelman(int **matrix, int m, int n);
int hoshen_kopelman(int ***matrix, int l, int m, int n);
int hoshen_kopelman(int **matrix, int **mask, int m, int n);
int hoshen_kopelman(int ***matrix, int ***mask, int l, int m, int n);
void uf_initialize(int max_labels);
int uf_make_set(void);
int uf_union(int x, int y); 
int uf_find(int x);
void uf_done(void);

int *labels;
int  n_labels = 0;     /* length of the labels array */

image_cc *Cluster(image_cc *image, int *nr)
{
      image_cc *label;
      int carry;
      
      if(strcmp(image->itype,"btd")==0)
      {
	int l,m,n;
	int ***matrix;
	
	l=image->dim[2];
	m=image->dim[1];
	n=image->dim[0];
	
	  matrix = (int ***)calloc(l, sizeof(int**));
	  for (int h=0; h<l; h++)
	  {
	    matrix[h] = (int **)calloc(m, sizeof(int*));
	    for (int i=0; i<m; i++) 
	    {
	      matrix[h][i] = (int *)calloc(n, sizeof(int));
	      for (int j=0; j<n; j++)
	      {
		matrix[h][i][j]=rPixel(j,i,h,image);
	      }
	    }
	  }
	  
	  
	  *nr = hoshen_kopelman(matrix,l,m,n);
	  
	  printf("HK reports %d clusters found\n", *nr);

	  if(*nr>=65536)
	      printf("Warning: Only 65535 clusters are written to label image. Use RemoveObjects() and LogAnd() in advance and start again to label them all\n");
	  label=InitImage(n,m,l,0,"ddd16");
	  for (int h=0; h<l; h++)
	    for (int i=0; i<m; i++) 
	      for (int j=0; j<n; j++)
	      {
		if((carry=matrix[h][i][j])<65536)wPixel(j,i,h,label,carry);
	      }
		

    
      }
      else if(strcmp(image->itype,"gry")==0)
      {
	int **matrix;
	int m,n;
	
	m=image->dim[1];
	n=image->dim[0];
	
	  
	  matrix = (int **)calloc(m, sizeof(int*));
	  
	  for (int i=0; i<m; i++) {
	    matrix[i] = (int *)calloc(n, sizeof(int));
	    for (int j=0; j<n; j++)
	      matrix[i][j]=rPixel(j,i,image);
	  }
	  
	  
	  *nr = hoshen_kopelman(matrix,m,n);
	  

	  
	  //check_labelling(matrix,m,n);
	  
	  printf("HK reports %d clusters found\n", *nr);

	  label=InitImage(n,m,0,"gry16");
	  for (int i=0; i<m; i++) {
	    for (int j=0; j<n; j++)
	      if((carry=matrix[i][j]%65536)!=0)wPixel(j,i,label,carry);
	  }
	  
	 
	  
	  for (int i=0; i<m; i++)
	    free(matrix[i]);
	  free(matrix);
	
  	
      }
      else{
	fprintf(stderr,"Only btd-format or 2D-tiff image allowed\n");
	exit(1);
      }
      
      
      return label;
}
image_cc *Cluster(image_cc *image, image_cc *maskim, int *nr)
{
      image_cc *label;
      int carry;
      
      if(strcmp(image->itype,"btd")==0)
      {
	int l,m,n;
	int ***matrix, ***mask;
	
	l=image->dim[2];
	m=image->dim[1];
	n=image->dim[0];
	
	  matrix = (int ***)calloc(l, sizeof(int**));
	  mask = (int ***)calloc(l, sizeof(int**));
	  for (int h=0; h<l; h++)
	  {
	    matrix[h] = (int **)calloc(m, sizeof(int*));
	    mask[h] = (int **)calloc(m, sizeof(int*));
	    for (int i=0; i<m; i++) 
	    {
	      matrix[h][i] = (int *)calloc(n, sizeof(int));
	      mask[h][i] = (int *)calloc(n, sizeof(int));
	      for (int j=0; j<n; j++)
	      {
		matrix[h][i][j]=rPixel(j,i,h,image);
		mask[h][i][j]=rPixel(j,i,h,maskim);
	      }
	    }
	  }
	  
	  
	  *nr = hoshen_kopelman(matrix,mask,l,m,n);
	  
	  printf("HK reports %d clusters found\n", *nr);

	  label=InitImage(n,m,l,0,"ddd16");
	  for (int h=0; h<l; h++)
	    for (int i=0; i<m; i++) 
	      for (int j=0; j<n; j++)
	      	if((carry=matrix[h][i][j]%65536)!=0)wPixel(j,i,h,label,carry);
		
	      

    
      }
      else if(strcmp(image->itype,"gry")==0)
      {
	int **matrix, **mask;
	int m,n;
	
	m=image->dim[1];
	n=image->dim[0];
	
	  
	  matrix = (int **)calloc(m, sizeof(int*));
	  mask = (int **)calloc(m, sizeof(int*));
	  
	  for (int i=0; i<m; i++) {
	    matrix[i] = (int *)calloc(n, sizeof(int));
	    mask[i] = (int *)calloc(n, sizeof(int));
	    for (int j=0; j<n; j++)
	    {
	      matrix[i][j]=rPixel(j,i,image);
	      mask[i][j]=rPixel(j,i,maskim);
	    }
	  }
	  
	  
	  *nr = hoshen_kopelman(matrix,mask,m,n);
	  

	  
	  //check_labelling(matrix,m,n);
	  
	  printf("HK reports %d clusters found\n", *nr);

	  label=InitImage(n,m,0,"gry16");
	  for (int i=0; i<m; i++) {
	    for (int j=0; j<n; j++)
	      if((carry=matrix[i][j]%65536)!=0)wPixel(j,i,label,carry);
	  }
	  
	 
	  
	  for (int i=0; i<m; i++)
	    free(matrix[i]);
	  free(matrix);
	
  	
      }
      else{
	fprintf(stderr,"Only btd-format or 2D-tiff image allowed\n");
	exit(1);
      }
      
      
      return label;
}
int hoshen_kopelman(int ***matrix, int l, int m, int n) {
  
  uf_initialize(l * m * n / 2); //is the worst case maximum cluster number on a checkerbord
  
  /* scan the matrix */
  for (int h=0; h<l; h++)  
    for (int i=0; i<m; i++)
      for (int j=0; j<n; j++)
	if (matrix[h][i][j]) 
	{                        // if occupied ...

	  int up = (h==0 ? 0 : matrix[h-1][i][j]);    //  look up
	  int back = (i==0 ? 0 : matrix[h][i-1][j]);    //  look back  
	  int left = (j==0 ? 0 : matrix[h][i][j-1]);  //  look left
	  
	  switch ((up ? 4:0) + (back ? 2:0) + (left ? 1:0)) 
	  {
	    
	  case 0:
	    matrix[h][i][j] = uf_make_set();      // a new cluster
	    break;
	  case 1:                              // part of an existing cluster
	    matrix[h][i][j] = left;       
	    break;
	  case 2:                              // part of an existing cluster
	    matrix[h][i][j] = back;       
	    break;
	  case 3:                              // this site binds two clusters
	    matrix[h][i][j] = uf_union(back, left);
	    break;	
	  case 4:                              // part of an existing cluster
	    matrix[h][i][j] = up;       
	    break;
	  case 5:                              // this site binds two clusters
	    matrix[h][i][j] = uf_union(up, left);
	    break;
	  case 6:                              // this site binds two clusters
	    matrix[h][i][j] = uf_union(up, back);
	    break;	    
	  case 7:                              // this site binds three clusters (only hypothetical)
	    int idum=uf_union(up, back);
	    matrix[h][i][j] = uf_union(idum, left);
	    break;
	  }
	  
	}
    
  /* apply the relabeling to the matrix */

  /* This is a little bit sneaky.. we create a mapping from the canonical labels
     determined by union/find into a new set of canonical labels, which are 
     guaranteed to be sequential. */
  int *new_labels;
  new_labels =(int*) calloc(sizeof(int), n_labels); // allocate array, initialized to zero
  for (int h=0; h<l; h++)
    for (int i=0; i<m; i++)
      for (int j=0; j<n; j++)
	if (matrix[h][i][j]) {
	  int x = uf_find(matrix[h][i][j]);
	  if (new_labels[x] == 0) {
	    new_labels[0]++;
	    new_labels[x] = new_labels[0];
	  }
	  matrix[h][i][j] = new_labels[x];
	}
 
  int total_clusters = new_labels[0];

  free(new_labels);
  uf_done();

  return total_clusters;
}

int hoshen_kopelman(int ***matrix, int ***mask, int l, int m, int n) {
  
  uf_initialize(l * m * n / 2); //is the worst case maximum cluster number on a checkerbord
  
  /* scan the matrix */
  for (int h=0; h<l; h++)  
    for (int i=0; i<m; i++)
      for (int j=0; j<n; j++)
	if (matrix[h][i][j] && mask[h][i][j]) 
	{                        // if occupied and within mask

	  int up = (h==0 ? 0 : matrix[h-1][i][j]);    //  look up
	  int back = (i==0 ? 0 : matrix[h][i-1][j]);    //  look back  
	  int left = (j==0 ? 0 : matrix[h][i][j-1]);  //  look left
	  
	  switch ((up ? 4:0) + (back ? 2:0) + (left ? 1:0)) 
	  {
	    
	  case 0:
	    matrix[h][i][j] = uf_make_set();      // a new cluster
	    break;
	  case 1:                              // part of an existing cluster
	    matrix[h][i][j] = left;       
	    break;
	  case 2:                              // part of an existing cluster
	    matrix[h][i][j] = back;       
	    break;
	  case 3:                              // this site binds two clusters
	    matrix[h][i][j] = uf_union(back, left);
	    break;	
	  case 4:                              // part of an existing cluster
	    matrix[h][i][j] = up;       
	    break;
	  case 5:                              // this site binds two clusters
	    matrix[h][i][j] = uf_union(up, left);
	    break;
	  case 6:                              // this site binds two clusters
	    matrix[h][i][j] = uf_union(up, back);
	    break;	    
	  case 7:                              // this site binds three clusters (only hypothetical)
	    int idum=uf_union(up, back);
	    matrix[h][i][j] = uf_union(idum, left);
	    break;
	  }
	  
	}
    
  /* apply the relabeling to the matrix */

  /* This is a little bit sneaky.. we create a mapping from the canonical labels
     determined by union/find into a new set of canonical labels, which are 
     guaranteed to be sequential. */
  int *new_labels;
  new_labels =(int*) calloc(sizeof(int), n_labels); // allocate array, initialized to zero
  for (int h=0; h<l; h++)
    for (int i=0; i<m; i++)
      for (int j=0; j<n; j++)
	if (matrix[h][i][j] && mask[h][i][j]) {
	  int x = uf_find(matrix[h][i][j]);
	  if (new_labels[x] == 0) {
	    new_labels[0]++;
	    new_labels[x] = new_labels[0];
	  }
	  matrix[h][i][j] = new_labels[x];
	}
 
  int total_clusters = new_labels[0];

  free(new_labels);
  uf_done();

  return total_clusters;
}

int hoshen_kopelman(int **matrix, int m, int n) {
  
  uf_initialize(m * n / 2);
  
  /* scan the matrix */
  
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      if (matrix[i][j]) {                        // if occupied ...
	int up = (i==0 ? 0 : matrix[i-1][j]);    //  look up  
	int left = (j==0 ? 0 : matrix[i][j-1]);  //  look left
	
	switch ((up ? 1:0) + (left ? 1:0)) {
	  
	case 0:
	  matrix[i][j] = uf_make_set();      // a new cluster
	  break;
	  
	case 1:                              // part of an existing cluster
	  matrix[i][j] = MAX(up,left);       // whichever is nonzero is labelled
	  break;
	  
	case 2:                              // this site binds two clusters
	  matrix[i][j] = uf_union(up, left);
	  //if(i==1 && j==1)
	   // printf("%d\t%d\t%d\n",matrix[i][j],uf_find(up),uf_find(left));
	  break;
	}
	
      }
  
  /* apply the relabeling to the matrix */

  /* This is a little bit sneaky.. we create a mapping from the canonical labels
     determined by union/find into a new set of canonical labels, which are 
     guaranteed to be sequential. */
  int *new_labels;
  new_labels =(int*) calloc(sizeof(int), n_labels); // allocate array, initialized to zero
  
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      if (matrix[i][j]) {
	if(i==1 && j==1)
	  printf("stop\n");
	int x = uf_find(matrix[i][j]);
	if (new_labels[x] == 0) {
	  new_labels[0]++;
	  new_labels[x] = new_labels[0];
	}
	matrix[i][j] = new_labels[x];
      }
 
  int total_clusters = new_labels[0];

  free(new_labels);
  uf_done();

  return total_clusters;
}

int hoshen_kopelman(int **matrix, int **mask, int m, int n) {
  
  uf_initialize(m * n / 2);
  
  /* scan the matrix */
  
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      if (matrix[i][j] && mask[i][j]) {                        // if occupied and within mask
	int up = (i==0 ? 0 : matrix[i-1][j]);    //  look up  
	int left = (j==0 ? 0 : matrix[i][j-1]);  //  look left
	
	switch ((up ? 1:0) + (left ? 1:0)) {
	  
	case 0:
	  matrix[i][j] = uf_make_set();      // a new cluster
	  break;
	  
	case 1:                              // part of an existing cluster
	  matrix[i][j] = MAX(up,left);       // whichever is nonzero is labelled
	  break;
	  
	case 2:                              // this site binds two clusters
	  matrix[i][j] = uf_union(up, left);
	  //if(i==1 && j==1)
	   // printf("%d\t%d\t%d\n",matrix[i][j],uf_find(up),uf_find(left));
	  break;
	}
	
      }
  
  /* apply the relabeling to the matrix */

  /* This is a little bit sneaky.. we create a mapping from the canonical labels
     determined by union/find into a new set of canonical labels, which are 
     guaranteed to be sequential. */
  int *new_labels;
  new_labels =(int*) calloc(sizeof(int), n_labels); // allocate array, initialized to zero
  
  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++)
      if (matrix[i][j] && mask[i][j]) {
	if(i==1 && j==1)
	  printf("stop\n");
	int x = uf_find(matrix[i][j]);
	if (new_labels[x] == 0) {
	  new_labels[0]++;
	  new_labels[x] = new_labels[0];
	}
	matrix[i][j] = new_labels[x];
      }
 
  int total_clusters = new_labels[0];

  free(new_labels);
  uf_done();

  return total_clusters;
}

void uf_initialize(int max_labels) {
  n_labels = max_labels;
  labels = (int*)calloc(sizeof(int), n_labels);
  labels[0] = 0;
}

/*  uf_make_set creates a new equivalence class and returns its label */

int uf_make_set(void) {
  labels[0] ++;
  assert(labels[0] < n_labels); // labels[0] (=number of labels) has to be smaller then array length, otherwise warning in debug mode
  labels[labels[0]] = labels[0]; // assign new label counter to new cluster
  return labels[0];
}

/*  uf_union joins two equivalence classes and returns the canonical label of the resulting class. */

int uf_union(int x, int y) {
  return labels[uf_find(x)] = uf_find(y);
}


/*  uf_find returns the canonical label for the equivalence class containing x */

int uf_find(int x) {
  int y = x;
  while (labels[y] != y)
    y = labels[y];
  
  while (labels[x] != x) {
    int z = labels[x];
    labels[x] = y;
    x = z;
  }
  return y;
}

/*  uf_done frees the memory used by the union-find data structures */

void uf_done(void) {
  n_labels = 0;
  free(labels);
  labels = 0;
}


