/*---------------------------------------------------------------------
VERSION 4.01,COPYRIGHT 2013 S. SCHLÜTER
quantim IS LICENSED FREE OF CHARGE. THEREFORE THIS FILE,AND ALL THE ACCOMPANYING FILES WHICH IN THEIR ENTIRETY CONSTITUTE quantim, ARE PROVIDED "AS IS",WITHOUT WARRANTY OF ANY KIND,WHETHER EXPRESSED OR IMPLIED. YOU ARE RESPONSIBLE FOR
ASCERTAINING THE FITNES
---------------------------------------------------------------------------*/


/**********************************************************************************************
FUNCTION: threshold detection according to Schlueter et al. (2010)

OUTPUT: pointer to integer vector
	t[0]=lower threshold
	t[1]=upper threshold

STEPS: 
1. load the image that has to be evaluated and convert it into an grey image if it has RGB format
2. create the gradient images (Sobel image first)
3. create gradient masks by simple thresholding of the gradient images via triangulation of their histograms according to Rosin(2001) - Sobel mask with one erosion
4. multiply the uploaded image by the gradient mask - store the intermediate images for visual control of performance
5. Repeat 3.-5. step with Laplace image (instead of Sobel)
6. calculate the upper threshold via weighted mean within the edge regions and lower threshold as a simple relation of the upper threshold and the image mode
7. write all thresholds into  thresh.txt

***********************************************************************************************/


#include "quantim4.h"
#define AVERAGE 3 /*range of running mean for mode-detection*/

int findmean_weighted(image_cc *origcond, image_cc *gradient);
int findmode(image_cc *image);
int findmode(image_cc *image, image_cc *mask);



int *GradMaskThresh(image_cc *image, double alpha)
	{
	image_cc *laplace, *sobel, *imagecondlap, *imagecondsob, *sobmask, *sobmaskerod, *lapmask, *pore, *copy;
	char buf[127];
	int maxdevsob, maxdevlap, xmode, is3d=1, is16bit=1,j,size,tlow,thigh, *t;
	double *laphisto, *condsobhisto, *condlaphisto, *sobhisto,dummy, lapthigh_weighted, sobthigh_weighted, *histo;
	FILE *fp;


/*********************************************
		1.STEP
**********************************************/
	if(image->numchannel!=1){ // tiff-files often have 3 channels even if they appear grey
		printf("this RGB image is converted into a one channel grey image\n");
		image_cc *dummyimage;
		dummyimage=CopyImage(image);
		image=RGBtoGray(dummyimage);
		DeleteImage(dummyimage);
        }

	if(image->ndim==3) is3d=image->dim[2];
	if(image->numbytes==2) is16bit=256;
	size=image->dim[0]*image->dim[1]*is3d; /* number of pixels for sequential for-loops through the image instead of nested dimensional loops*/
	

	
/*********************************************
		2.STEP
**********************************************/

	printf("generating sobel image (1. derivative of gray values)\n");
	sobel=Sobel(image);
	sprintf(buf,"sobel");
	StoreImage(sobel,buf);

/*********************************************
		3.STEP
**********************************************/	
	sobhisto=Histo(sobel);

	maxdevsob=UnimThresh(sobhisto,1);
	maxdevsob*=is16bit;
	if(sobel->ndim==3) sobmask=DddBin(sobel,maxdevsob);
	else{
		sobmask=CopyImage(sobel);
		Bin(sobmask,maxdevsob);
	}
	
	sobmaskerod=ErodeCirc(sobmask, 1, 1);
	//sobmaskerod=CopyImage(sobmask);
	DeleteImage(sobmask);


/*********************************************
		4.STEP
**********************************************/	

	xmode=findmode(image);/* determine the image mode to exlude edges to bright objects from evaluation*/ 
	printf("image mode: %d\n", xmode);
	

	/* to create conditional images the non-edge pixels and bright-edge pixels are written to zero */	
	imagecondsob=CopyImage(image);
	for(j=0; j<size;j++) if( (rPixel(j,image)>xmode) || (!rPixel(j,sobmaskerod))) wPixel(j,imagecondsob,0);
	DeleteImage(sobmaskerod);

	/* store the conditional images*/	
	sprintf(buf, "sobcond");
	StoreImage(imagecondsob, buf);

	condsobhisto=Histo(imagecondsob);
	sobthigh_weighted=findmean_weighted(imagecondsob, sobel);
	DeleteImage(imagecondsob);
	DeleteImage(sobel);
	
	

/*********************************************
		5.STEP
**********************************************/	
	
	
	printf("generating laplace image (2. derivative of gray values)\n"); 
	laplace=Laplace(image,1);
	sprintf(buf,"laplace");
	StoreImage(laplace,buf);
	laphisto=Histo(laplace);
	maxdevlap=UnimThresh(laphisto,1);
	maxdevlap*=is16bit;
	
	if(laplace->ndim==3) lapmask=DddBin(laplace,maxdevlap);
	else{
		lapmask=CopyImage(laplace);
		Bin(lapmask,maxdevlap);
	}
	
	imagecondlap=CopyImage(image);
	for(j=0; j<size;j++) if( (rPixel(j,image)>xmode) || (!rPixel(j,lapmask))) wPixel(j,imagecondlap,0);
	DeleteImage(lapmask);
	DeleteImage(image);
	sprintf(buf, "lapcond");
	StoreImage(imagecondlap, buf);
	condlaphisto=Histo(imagecondlap);
	lapthigh_weighted=findmean_weighted(imagecondlap, laplace);
	DeleteImage(imagecondlap);
	DeleteImage(laplace);
	
/*********************************************
		6.STEP
**********************************************/
	thigh=int((sobthigh_weighted+lapthigh_weighted)/2);
	tlow=int(alpha*thigh+(1-alpha)*xmode);

/*********************************************
		7.STEP
**********************************************/
	printf("see thresh.txt for histogramm-statistics\n\n");
	sprintf(buf,"thresh.txt");
	if ( (fp = fopen(buf, "w")) == NULL)
	{
	    fprintf(stderr, "Fehler beim Oeffnen der Datei %s.\n", buf);
	    exit(1);
	}
	fprintf(fp,"thresh.txt\n");
	fprintf(fp,"Histogramm-Mode\t%d\n", xmode);
	fprintf(fp,"Laplace-Mask-Mean\t%d\n", int(lapthigh_weighted));
	fprintf(fp,"Sobel-Mask-Mean\t%d\n", int(sobthigh_weighted));
	fprintf(fp,"Thigh(2Mask-Mean)\t%d\n",thigh);
	fprintf(fp,"Faktor\t%f\n", alpha);
	fprintf(fp,"Tlow=\t%d\n", tlow);
	fclose(fp);
	
	t=(int*)malloc(2*sizeof(int));
	t[0]=tlow;
	t[1]=thigh;
	return t;
}

int *GradMaskThresh(image_cc *image, double alpha, image_cc *mask)
	{
	image_cc *laplace, *sobel, *imagecondlap, *imagecondsob, *sobmask, *sobmaskerod, *lapmask, *pore, *copy;
	char buf[127];
	int maxdevsob, maxdevlap, xmode, is3d=1, is16bit=1,j,size,tlow,thigh, *t;
	double *laphisto, *condsobhisto, *condlaphisto, *sobhisto,dummy, lapthigh_weighted, sobthigh_weighted, *histo;
	FILE *fp;


/*********************************************
		1.STEP
**********************************************/
	if(image->numchannel!=1){ // tiff-files often have 3 channels even if they appear grey
		printf("this RGB image is converted into a one channel grey image\n");
		image_cc *dummyimage;
		dummyimage=CopyImage(image);
		image=RGBtoGray(dummyimage);
		DeleteImage(dummyimage);
        }

	if(image->ndim==3) is3d=image->dim[2];
	if(image->numbytes==2) is16bit=256;
	size=image->dim[0]*image->dim[1]*is3d; /* number of pixels for sequential for-loops through the image instead of nested dimensional loops*/
	
	
/*********************************************
		2.STEP
**********************************************/

	printf("generating sobel image (1. derivative of gray values)\n");
	sobel=Sobel(image);
	sprintf(buf,"sobel");
	StoreImage(sobel,buf);

/*********************************************
		3.STEP
**********************************************/	
	sobhisto=Histo(sobel);

	maxdevsob=UnimThresh(sobhisto,1);
	maxdevsob*=is16bit;
	if(sobel->ndim==3) sobmask=DddBin(sobel,maxdevsob);
	else{
		sobmask=CopyImage(sobel);
		Bin(sobmask,maxdevsob);
	}
	
	sobmaskerod=ErodeCirc(sobmask, 1, 1);
	//sobmaskerod=CopyImage(sobmask);
	DeleteImage(sobmask);


/*********************************************
		4.STEP
**********************************************/	

	xmode=findmode(image,mask);/* determine the image mode to exlude edges to bright objects from evaluation*/ 
	printf("image mode: %d\n", xmode);
	

	/* to create conditional images the non-edge pixels and bright-edge pixels are written to zero */	
	imagecondsob=CopyImage(image);
	for(j=0; j<size;j++) if( (rPixel(j,image)>xmode) || (!rPixel(j,sobmaskerod))) wPixel(j,imagecondsob,0);
	DeleteImage(sobmaskerod);

	/* store the conditional images*/	
	sprintf(buf, "sobcond");
	StoreImage(imagecondsob, buf);

	condsobhisto=Histo(imagecondsob);
	sobthigh_weighted=findmean_weighted(imagecondsob, sobel);
	DeleteImage(imagecondsob);
	DeleteImage(sobel);
	
	

/*********************************************
		5.STEP
**********************************************/	
	
	
	printf("generating laplace image (2. derivative of gray values)\n"); 
	laplace=Laplace(image,1);
	sprintf(buf,"laplace");
	StoreImage(laplace,buf);
	laphisto=Histo(laplace);
	maxdevlap=UnimThresh(laphisto,1);
	maxdevlap*=is16bit;
	
	if(laplace->ndim==3) lapmask=DddBin(laplace,maxdevlap);
	else{
		lapmask=CopyImage(laplace);
		Bin(lapmask,maxdevlap);
	}
	
	imagecondlap=CopyImage(image);
	for(j=0; j<size;j++) if( (rPixel(j,image)>xmode) || (!rPixel(j,lapmask))) wPixel(j,imagecondlap,0);
	DeleteImage(lapmask);
	DeleteImage(image);
 	sprintf(buf, "lapcond");
 	StoreImage(imagecondlap, buf);
	condlaphisto=Histo(imagecondlap);
	lapthigh_weighted=findmean_weighted(imagecondlap, laplace);
	DeleteImage(imagecondlap);
	DeleteImage(laplace);
	
/*********************************************
		6.STEP
**********************************************/
	thigh=int((sobthigh_weighted+lapthigh_weighted)/2);
	tlow=int(alpha*thigh+(1-alpha)*xmode);

/*********************************************
		7.STEP
**********************************************/
	printf("see thresh.txt for histogramm-statistics\n\n");
	sprintf(buf,"thresh.txt");
	if ( (fp = fopen(buf, "w")) == NULL)
	{
	    fprintf(stderr, "Fehler beim Oeffnen der Datei %s.\n", buf);
	    exit(1);
	}
	fprintf(fp,"thresh.txt\n");
	fprintf(fp,"Histogramm-Mode\t%d\n", xmode);
	fprintf(fp,"Laplace-Mask-Mean\t%d\n", int(lapthigh_weighted));
	fprintf(fp,"Sobel-Mask-Mean\t%d\n", int(sobthigh_weighted));
	fprintf(fp,"Thigh(2Mask-Mean)\t%d\n",thigh);
	fprintf(fp,"Faktor\t%f\n", alpha);
	fprintf(fp,"Tlow=\t%d\n", tlow);
	fclose(fp);
	
	t=(int*)malloc(2*sizeof(int));
	t[0]=tlow;
	t[1]=thigh;
	return t;
}



int findmode(image_cc *image){
/*FUNCTION: Finds mode in 16bit histogram (smoothed by moving mean). This is more exect than previously converting into 8bit (as done in Histo() ) */
	
	int count, count2,  avmode, is3d=1,is16bit=1, size,j, window;
	if(image->ndim==3) is3d=image->dim[2];
	size=image->dim[0]*image->dim[1]*is3d;
	if(image->numbytes==2){
		is16bit=256;
		window=AVERAGE;
	} 
	else window=0;

	double dummy, histo[256*is16bit], average[256*is16bit], sum=0, maskpart;

	/*INITIALIZE*/
	for(count=0; count<256*is16bit; count++){
		histo[count]=0;
		average[count]=0;
	}

	/*ORIGINAL HISTOGRAM IN 16BIT*/
	for(j=0; j<size; j++){
		histo[rPixel(j,image)]++;
	}
	/*SET BLACK MASK TO ZERO*/	
	maskpart=histo[0]/double(size);
	histo[0]=0;
	
	/*MOVING MEAN ON HISTOGRAM*/
	for(count=window/2; count<256*is16bit-window/2; count++){
		for(count2=-window/2; count2<=window/2; count2++){
			  average[count]+=histo[count+count2];  
			  }
		average[count]/=(window+1);
	}

	/*NORMALIZE SMOOTHED HISTOGRAM*/
	for(count=0; count<256*is16bit; count++){
		average[count]/=((1-maskpart)*size);
		sum+=average[count];
	}
	//printf("sum: %lf\n", sum);


	/*FIND MODE IN SMOOTHED HISTOGRAM*/
	dummy=avmode=0;
	for(count=window/2; count<256*is16bit-window/2; count++){
		if (average[count]>dummy){
			dummy=average[count]; 
			avmode=count; //Modus
		}
		
	}
	
	return(avmode);
}
int findmode(image_cc *image, image_cc *mask){
	
	int count, count2, avmode, is3d=1,is16bit=1, size,j, window,roi=0;
	if(image->ndim==3) is3d=image->dim[2];
	size=image->dim[0]*image->dim[1]*is3d;
	if(image->numbytes==2){
		is16bit=256;
		window=AVERAGE;
	} 
	else window=0;

	double dummy, histo[256*is16bit], average[256*is16bit], sum=0, maskpart;
	
	for(count=0; count<256*is16bit; count++){
		histo[count]=0;
		average[count]=0;
	}

	for(j=0; j<size; j++)
	{
 		if(rPixel(j,mask))
		{
		  histo[rPixel(j,image)]++;
		  roi++;
		}
	}

	for(count=window/2; count<256*is16bit-window/2; count++){
		for(count2=-window/2; count2<=window/2; count2++){
			  average[count]+=histo[count+count2];  
			  }
		average[count]/=(window+1);
	}


	for(count=0; count<256*is16bit; count++){
		average[count]/=double(roi);
		sum+=average[count];
	}
	//printf("sum: %lf\n", sum);


	dummy=avmode=0;
	for(count=window/2; count<256*is16bit-window/2; count++){
		if (average[count]>dummy){
			dummy=average[count]; 
			avmode=count; //Modus
		}
		
	}
	
	return(avmode);
}
int findmean_weighted(image_cc *origcond, image_cc *gradient){

	image_cc *weighted;
	int size,j,is3d=1,is16bit=1;
	double total=0, sum=0; 
	char buf[127];

	if(origcond->ndim==3) is3d=origcond->dim[2];
	if(origcond->numbytes==2) is16bit=256;
	size=origcond->dim[0]*origcond->dim[1]*is3d; /* number of pixels for sequential for-loops through the image instead of nested dimensional loops*/

	weighted=CopyImage(origcond);
	for(j=0;j<size;j++){
		wPixel(j,weighted,int(rPixel(j,origcond)*rPixel(j,gradient)/(256*is16bit)));/*product may cause 16bit range overflow. Therefore division with large number*/
		if(rPixel(j,origcond)) sum+=double(rPixel(j,gradient))/double(size); /*prevent sum overflow*/
	}
	/*sprintf(buf,"weighted");
	StoreImage(weighted,buf);*/ //Map of weighted gradients
	for(j=0;j<size;j++){
		total+=double(rPixel(j,weighted))/double(sum*size)*256*double(is16bit); /*sum of corrected weighting factors has to be one*/
				  
	}
	DeleteImage(weighted);
	//printf("weighted mean\t%f\n",total);
	return (total);
}