#include "quantim4.h"
#include <time.h>

/*--------------------------------------------------------------------------------------
grf2: generate random function in 2d with expectation 0 and variance VAR
VARIABLES:
  sfa		vector to return data (cut from larger field)
  DX,DY		spatial discretisation
  LAMX,LAMY	correlation lengths
  VAR		variance 
  N		# nodes in one direction (must be power of 2)
  TYPE:		0: Gauss, 1: Lorentz, 2: exponetial, 3: von Karman)
--------------------------------------------------------------------------------------*/  

void grf2(double*sfa,int NX,int NY,double DX,double DY,double LAMX,double LAMY,double VAR,int N,int TYPE)
{
  unsigned long dim[3];
  int i,j,l,n,m;
  long idum; 
  float *buf,s1,s2,*phi,r,r0;

  buf=(float*)malloc(int((2.*pow(N+1,2)))*sizeof(float)); buf--;
  phi=(float*)malloc(int(pow(N+1,2))*sizeof(float));
  time(&idum);
  ran3(&idum);				/* initialize random generator */
  dim[1]=dim[2]=N;
  for (i=0;i<pow(N+1,2);i++) phi[i]=2*M_PI*ran3(&idum);

  if (TYPE==0) {				/* Gauss */
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=exp(-0.25*M_PI*(pow(i*DX/LAMX,2)+pow(j*DY/LAMY,2)));
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }
  if (TYPE==1) {                              /* Lorentz */
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=1/(1+pow((M_PI/2),2)*(pow(i*DX/LAMX,2)+pow(j*DY/LAMY,2)));
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }
  if (TYPE==2) {                              /* Exponential */
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=exp(-sqrt(pow(i*DX/LAMX,2)+pow(j*DY/LAMY,2)));
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }
  if (TYPE==3) {                              /* von Karman */
    r0=0.0001*DX*DY/(LAMX*LAMY);
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=sqrt(pow(i*DX/LAMX,2)+pow(j*DY/LAMY,2)); if (r<r0) r=r0;
	r=(2/M_PI)*K0(r);
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }

  fourn(buf,dim,2,1);
  for (i=0;i<N*(N+1);i++) {
    r=pow(pow(buf[2*i+1],2)+pow(buf[2*(i+1)],2),0.25);
    buf[2*i+1]=r*cos(phi[i]); buf[2*(i+1)]=r*sin(phi[i]);
  }
  fourn(buf,dim,2,-1);
						/* correct statistics of realization */
  s1=0; for (i=0;i<NX;i++) for (j=0;j<NY;j++) s1+=buf[2*(i+j*N)+1];
  s1/=NX*NY;
  s2=0; for (i=0;i<NX;i++) for (j=0;j<NY;j++) s2+=pow(buf[2*(i+j*N)+1]-s1,2);
  s2=sqrt(NX*NY*VAR/s2);
  for (i=0;i<NX;i++) for (j=0;j<NY;j++) sfa[i+j*NX]=s2*(buf[2*(i+j*N)+1]-s1);

  free((void*)++buf); free((void*)phi);
}
 


void grf2ext(double*sfa,int NX,int NY,double DX,double DY,double LAMX,double LAMY,double VAR,int N,int TYPE, double **acov)
{
  unsigned long dim[3];
  int i,j,l,n,m;
  long idum; 
  float *buf,s1,s2,*phi,r,r0;

  buf=(float*)malloc(int((2.*pow(N+1,2)))*sizeof(float)); buf--;
  phi=(float*)malloc(int(pow(N+1,2))*sizeof(float));
  time(&idum);
  ran3(&idum);				/* initialize random generator */
  dim[1]=dim[2]=N;
  for (i=0;i<pow(N+1,2);i++) phi[i]=2*M_PI*ran3(&idum);

  if (TYPE==0) {				/* Gauss */
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=exp(-0.25*M_PI*(pow(i*DX/LAMX,2)+pow(j*DY/LAMY,2)));
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }
  if (TYPE==1) {                              /* Lorentz */
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=1/(1+pow((M_PI/2),2)*(pow(i*DX/LAMX,2)+pow(j*DY/LAMY,2)));
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }
  if (TYPE==2) {                              /* Exponential */
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=exp(-sqrt(pow(i*DX/LAMX,2)+pow(j*DY/LAMY,2)));
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }
  if (TYPE==3) {                              /* von Karman */
    r0=0.0001*DX*DY/(LAMX*LAMY);
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=sqrt(pow(i*DX/LAMX,2)+pow(j*DY/LAMY,2)); if (r<r0) r=r0;
	r=(2/M_PI)*K0(r);
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }
  if (TYPE==4) {                              /* predefined  */
    for (i=0;i<=N/2;i++) {
      for (j=0;j<=N/2;j++) {
        r=acov[0][i]*DX*acov[1][j]*DY;
	m= i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n= j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m= i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
	m=-i+N/2; n=-j+N/2; l=2*(m+n*N)+1; buf[l]=r; buf[l+1]=0;
      }
    }
  }

  fourn(buf,dim,2,1);
  for (i=0;i<N*(N+1);i++) {
    r=pow(pow(buf[2*i+1],2)+pow(buf[2*(i+1)],2),0.25);
    buf[2*i+1]=r*cos(phi[i]); buf[2*(i+1)]=r*sin(phi[i]);
  }
  fourn(buf,dim,2,-1);
						/* correct statistics of realization */
  s1=0; for (i=0;i<NX;i++) for (j=0;j<NY;j++) s1+=buf[2*(i+j*N)+1];
  s1/=NX*NY;
  s2=0; for (i=0;i<NX;i++) for (j=0;j<NY;j++) s2+=pow(buf[2*(i+j*N)+1]-s1,2);
  s2=sqrt(NX*NY*VAR/s2);
  for (i=0;i<NX;i++) for (j=0;j<NY;j++) sfa[i+j*NX]=s2*(buf[2*(i+j*N)+1]-s1);

  free((void*)++buf); free((void*)phi);
}
 

 

 
